package EqMath 
{
	/**
	 * This class is used to solve cubic equations
	 * @author Shiu
	 */
	public class EqCubic implements IEquation {
		public static var degree:int = 3;
		
		private var _A:Number, _B:Number, _C:Number, _D:Number;
		private var _roots_R:Vector.<Number>;
		private var _roots_i:Vector.<String>;
		
		//storage to optimise calculation
		private var p:Number, q:Number, r:Number;
		private var a:Number, b:Number;
		private var frag1:Number; //	a/3
		private var frag2:Number; //	b/2
		
		/**
		 * Note the order of degrees writen in formula template
		 * f(x) = A (x^3) + B (x^2) + C (x) + D
		 */
		public function EqCubic() {
			_roots_R = new Vector.<Number>;	_roots_R.length = 3;
			_roots_i = new Vector.<String>;	_roots_i.length = 3;
		}
		
		/**
		 * Define the coefficients of quadratic equation
		 * @param	...coeff Coefficients A, B, C, D
		 */
		public function define (...coeff):void {
			_A = coeff[0]; _B = coeff[1];  _C = coeff[2]; _D = coeff[3];
			
			//set up the initial values
			_roots_R[0] = Number.NaN; _roots_i[0] = null;
			_roots_R[1] = Number.NaN; _roots_i[1] = null
			_roots_R[2] = Number.NaN; _roots_i[2] = null; 
		}
		
		/**
		 * Get f(x) based upon x
		 * @param	x value on x-axis
		 * @return f(x) or value on y-axis
		 */
		public function fx_of(x:Number):Number {
			return _A * x * x * x + _B * x * x + _C * x + _D;
		}
		
		/****************** Begin: Solving for root ******************/
		
		/**Read-only
		 * Discriminant of equation
		 */
		public function get discriminant():Number {
			//precalculate of variables for solving root
			 p = _B / _A; 
			 q = _C / _A;
			 r = _D / _A;
			 
			 a = (3 * q - p * p) / 3;
			 b = (2 * p * p * p - 9 * p * q +27 * r) / 27;
			 
			 frag1 = a / 3;
			 frag2 = b / 2;
			 
			 //now the discriminant
			return frag2 * frag2 + frag1 * frag1 * frag1;
		}
		
		/**
		 * Performs calculation to obtain roots
		 */
		public function calcRoots():void {
			var disc:Number = this.discriminant
			
			var component_real:Number;
			var component_imaginary:Number;
			var sqrt_1:Number = Math.sqrt( -frag1);
			
			//when discriminant is more than zero,
			//one real root, two conjugate imaginary roots
			if (disc > 0) { 	
				
				//handle positive & negative cubic roots
				var b4_cbrt1:Number = -frag2 + Math.sqrt(disc);
				var b4_cbrt2:Number = -frag2 - Math.sqrt(disc);
				
				var cbrt1:Number, cbrt2:Number;
				if (b4_cbrt1 > 0) cbrt1 = Math.pow(b4_cbrt1, 1 / 3);
				else cbrt1 = -Math.pow( -b4_cbrt1, 1 / 3);
				
				if (b4_cbrt2 > 0) cbrt2 = Math.pow(b4_cbrt2, 1 / 3);
				else cbrt2 = -Math.pow( -b4_cbrt2, 1 / 3);
				
				//one real
				_roots_R[0] = cbrt1 + cbrt2;
				
				//two imaginaries
				component_real = (cbrt1 + cbrt2) * -0.5;	
				component_imaginary = Math.sqrt(3) * (cbrt1 - cbrt2) * 0.5;
				_roots_i[0] = (component_real + "+ i" + component_imaginary).toString();
				_roots_i[1] = (component_real + "- i" + component_imaginary).toString();
			}
			
			//when discriminant is equal to zero,
			//all roots are real, of which two are equal
			else if (disc == 0) {
				
				//handle b equals 0
				if (b == 0) {	
					_roots_R[0] = _roots_R[1] = _roots_R[2] = 0;
					return;
				}
				
				//handle b < 0 and b > 0;
				_roots_R[0] = 2 * sqrt_1;
				_roots_R[1] = -1 * sqrt_1;
				_roots_R[2] = -1 * sqrt_1;
				
				if (b > 0) {
					_roots_R[0] *= -1;
					_roots_R[1] *= -1;
					_roots_R[2] *= -1;
				}
			}
			
			//when discriminant is less than zero,
			//all roots are real and unequal
			else {
				var sqrt_2:Number = Math.sqrt(	(frag2 * frag2)	/	(frag1 * frag1 * frag1 * -1)	);
				if (b > 0) sqrt_2 *= -1;
				var angle:Number = Math.acos(sqrt_2);
				_roots_R[0] = 2 * sqrt_1 * Math.cos(angle / 3);
				_roots_R[1] = 2 * sqrt_1 * Math.cos((angle + Math.PI * 2) / 3);
				_roots_R[2] = 2 * sqrt_1 * Math.cos((angle + Math.PI * 4) / 3)
			}
			
			for (var i:int = 0; i < _roots_R.length; i++) {
				if (!isNaN(_roots_R[i]))	_roots_R[i] -= p / 3;
			}
		}
		/****************** End: Solving for root ******************/
		
		/**
		 * Returns real roots if any exists, else returns NaN
		 */
		public function get roots_R():Vector.<Number> {
			return _roots_R;
		}
		
		/**
		 * Returns imaginary roots if any exists, else returns null
		 */
		public function get roots_i():Vector.<String> {
			return _roots_i;
		}
		
		/**
		 * Return first degree differentiation
		 * @param	x the x coordinate
		 */
		public function diff1(x:Number):Number {
			/* f(x) = ax ^ 3 + bx ^2 + cx + d
			 * f'(x), first degree differentiation
			 * f'(x) = 3ax ^ 2 + 2bx + c
			 */
			return 3 * _A * x * x + 2 * _B * x + _C;
		}
		
		/**
		 * Return second degree differentiation
		 * @param	x the x coordinate
		 */
		public function diff2(x:Number):Number {
			/* f(x) = ax ^ 3 + bx ^2 + cx + d
			 * f''(X), second degree differentiation
			 * f''(x) = 6ax + 2b
			 */
			return 6 * _A * x + 2 * _B;
		}
	}

}